/*
@(#)File:           $RCSfile: basename.c,v $
@(#)Version:        $Revision: 2.5 $
@(#)Last changed:   $Date: 2018/01/29 22:14:30 $
@(#)Purpose:        Return basename of filename
@(#)Author:         J Leffler
@(#)Copyright:      (C) JLSS 1989-2018
@(#)Product:        SQLCMD Version 91.02 (2018-10-28)
*/

/*TABSTOP=4*/

#include "jlss.h"
#include <string.h>

#ifndef lint
/* Prevent over-aggressive optimizers from eliminating ID string */
extern const char jlss_id_basename_c[];
const char jlss_id_basename_c[] = "@(#)$Id: basename.c,v 2.5 2018/01/29 22:14:30 jonathanleffler Exp $";
#endif /* lint */

/*
** Note that this function takes a constant string so it does not modify
** it.  It returns the last component of the file name, but there will
** be trailing slashes if there were trailing slashes in the source.
** POSIX-standard basename() takes a non-const string and zaps trailing
** slashes.  The POSIX specification means it can't legitimately process
** string constants.
*/

const char *jlss_basename(const char *s)
{
    const char  *cp;

    /* POSIX requires null pointer and empty string to return "." */
    if (s == 0 || *s == '\0')
        return ".";

    /* Strip trailing slashes, worrying about "/" */
    if ((cp = strrchr(s, '/')) != 0 && cp > s && cp[1] == '\0')
    {
        /* Skip backwards over trailing slashes */
        while (cp > s && *cp == '/')
            cp--;
        /* Skip backwards over non-slashes */
        while (cp > s && *cp != '/')
            cp--;
    }

    return((cp == 0 || (cp == s && (cp[1] == '\0' || cp[1] == '/'))) ? s : cp + 1);
}

#ifdef TEST
#include <stdio.h>

struct TestCase
{
    const char *input;
    const char *output;
};

static const struct TestCase tests[] =
{
    { "/usr/fred/bloggs",     "bloggs"       },
    { "/usr/fred/bloggs/",    "bloggs/"      },
    { "/usr/fred/bloggs////", "bloggs////"   },
    { "bloggs",               "bloggs"       },
    { "/.",                   "."            },
    { "/..",                  ".."           },
    { "./",                   "./"           },
    { ".",                    "."            },
    { "..",                   ".."           },
    { "./..",                 ".."           },
    { "/",                    "/"            },
    { "//",                   "//"           },
    { "///",                  "///"          },
    { "////",                 "////"         },
    { "a",                    "a"            },
    { ".a",                   ".a"           },
    { "a/",                   "a/"           },
    { "  abc  def  ",         "  abc  def  " },
    { "",                     "."            },
    { 0,                      "."            },
};
enum { NUM_TEST_CASES = sizeof(tests) / sizeof(tests[0]) };

static void test_one(const struct TestCase *tc)
{
    const char *res = jlss_basename(tc->input);
    char buffer1[32];
    const char *in = tc->input;
    if (in == NULL)
        in = "<null pointer>";
    snprintf(buffer1, sizeof(buffer1), "'%s'", in);
    if (strcmp(tc->output, res) == 0)
        printf("** PASS ** %-22s -> '%s'\n", buffer1, res);
    else
    {
        char buffer2[32];
        snprintf(buffer2, sizeof(buffer2), "'%s'", tc->output);
        printf("== FAIL == %-22s -> wanted %-22s actual '%s'\n", buffer1, buffer2, res);
    }
}

int main(void)
{
    for (int i = 0; i < NUM_TEST_CASES; i++)
        test_one(&tests[i]);
    return(0);
}
#endif /* TEST */
