/*
@(#)File:           $RCSfile: sqlinfo.ec,v $
@(#)Version:        $Revision: 2017.7 $
@(#)Last changed:   $Date: 2017/11/27 06:39:37 $
@(#)Purpose:        Handle INFO statements
@(#)Author:         J Leffler
@(#)Copyright:      (C) JLSS 1998-2017
@(#)Product:        SQLCMD Version 91.02 (2018-10-28)
*/

/*TABSTOP=4*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "connect.h"
#include "context.h"
#include "esqlutil.h"
#include "internal.h"
#include "jlss.h"
#include "jlsstools.h"
#include "esnprintf.h"
#include "sqlerr.h"
#include "debug.h"
#include "emalloc.h"
#include "sqlownobj.h"

#ifndef lint
/* Prevent over-aggressive optimizers from eliminating ID string */
extern const char jlss_id_sqlinfo_ec[];
const char jlss_id_sqlinfo_ec[] = "@(#)$Id: sqlinfo.ec,v 2017.7 2017/11/27 06:39:37 jleffler Exp $";
#endif /* lint */

/*
** The alternative command execution functions for INFO statements are:
**  -- do_command():  provides benchmark, trace and history service.
**  -- sql_command(): does not provide benchmark, trace or history.
** There are advantages to each.  To make it easier to change our mind,
** use a function pointer rather than a fixed command name throughout.
** One more peculiar side-effect of using do_command() is that the
** history number jumps by two or more when an INFO command is executed.
** On the other hand, during development of INFO commands, it is helpful
** to be able to track the SQL that is executed.
*/
static void (*sql_exec)(const char *cmd) = do_command;

/* Statement for INFO TABLES */
/* SE uses TabID 100 with TabType = ' ' to record MODE ANSI status */
static char info_tables[] =
    "SELECT Owner, TabName, TabType, TabID FROM \"informix\".SysTables T\n"
    " WHERE TabID >= 100\n"
    "   AND TabType IN ('T', 'V', 'P', 'S')\n"
    "   AND (%.*s)\n"
    " ORDER BY TabName, Owner;";

/* Statement for INFO SYSTABLES */
static char info_systables[] =
    "SELECT Owner, TabName, TabType, TabID FROM \"informix\".SysTables T\n"
    " WHERE TabID < 100 AND TabType IN ('T', 'V', 'P', 'S')\n"
    "   AND TabName NOT MATCHES ' *'\n"
    "   AND (%.*s)\n"
    " ORDER BY TabName, Owner;";

/* Statement for INFO VIEWS */
static char info_views[] =
    "SELECT Owner, TabName, TabID FROM \"informix\".SysTables T\n"
    " WHERE TabID >= 100 AND TabType = 'V'\n"
    "   AND (%.*s)\n"
    " ORDER BY TabName, Owner;";

/* Statement for INFO BASETABLES */
static char info_basetables[] =
    "SELECT Owner, TabName, TabID FROM \"informix\".SysTables T\n"
    " WHERE TabID >= 100 AND TabType = 'T'\n"
    "   AND (%.*s)\n"
    " ORDER BY TabName, Owner;";

/* Statement for INFO SYNONYMS */
/* Only report the user's private synonyms (P) but all the public ones (S) */
static char info_synonyms[] =
    "SELECT Owner, TabName, TabType, TabID FROM \"informix\".SysTables T\n"
    " WHERE TabID >= 100 AND (%.*s)\n"
    "   AND ((TabType = 'S') OR -- Public synonyms\n"
    "        (TabType = 'P' AND Owner = USER)) -- Private synonyms for this user\n"
    " ORDER BY TabName, Owner;";

/* Statement for INFO COLUMNS */
static char sql_columns[] =
    "SELECT ColNo, ColName, ColType, ColLength, Extended_ID\n"
    "  FROM %s\"informix\".SysColumns C\n"
    " WHERE TabID = %ld ORDER BY ColNo;";

/* Statement for INFO PROCEDURES */
static char info_procedures[] =
    "SELECT ProcID, Owner, Procname, NumArgs\n"
    "  FROM \"informix\".SysProcedures P\n"
    " WHERE (%.*s)\n"
    " ORDER BY ProcID;";

/* Statement for INFO PROCBODY */
static char sql_procbody[] =
    "SELECT Seqno, Data\n"
    "  FROM %s\"informix\".SysProcBody\n"
    " WHERE ProcID = %ld\n"
    "   AND DataKey = 'T'\n"
    " ORDER BY SeqNo;";

/* Statement for INFO VIEWBODY */
static char sql_viewbody[] =
    "SELECT Seqno, ViewText\n"
    "  FROM %s\"informix\".SysViews\n"
    " WHERE TabID = %ld\n"
    " ORDER BY SeqNo;";

/* Statement for INFO TRIGGERS */
/* Two conditions on TabID is overkill */
static char sql_triggers[] =
    "SELECT R.Owner, R.TrigName, R.TrigID, T.Owner TabOwner, T.TabName, R.Event, R.Old, R.New, R.Mode\n"
    "  FROM %s\"informix\".SysTables T\n"
    "  JOIN %s\"informix\".SysTriggers R ON R.TabID = T.TabID\n"
    " WHERE R.TabID = %ld\n"
    "   AND T.TabID = %ld\n"
    " ORDER BY R.TrigName, R.Owner;";

/* Statement for INFO TRIGBODY */
/* Need DataKey D (trigger definition) before DataKey A (triggered actions) */
static char sql_trigbody[] =
    "SELECT DataKey, Seqno, Data\n"
    "  FROM %s\"informix\".SysTrigBody\n"
    " WHERE TrigID = %ld\n"
    "   AND DataKey IN ('D', 'A')\n"
    " ORDER BY DataKey DESC, SeqNo ASC;";

/* Statement for INFO REFERENCES TO/BY */
static char sql_references[] =
    "SELECT C1.constrname   refng_constr_name,\n"
    "       C1.owner        refng_constr_owner,\n"
    "       C1.idxname      refng_index_name,\n"
    "       T1.tabid        refng_table_tabid,\n"
    "       T1.owner        refng_table_owner,\n"
    "       T1.tabname      refng_table_name,\n"
    "       C2.constrname   refed_constr_name,\n"
    "       C2.owner        refed_constr_owner,\n"
    "       C2.idxname      refed_index_name,\n"
    "       T2.tabid        refed_table_tabid,\n"
    "       T2.owner        refed_table_owner,\n"
    "       T2.tabname      refed_table_name\n"
    "  FROM %s\"informix\".SysReferences R\n"
    "  JOIN %s\"informix\".SysConstraints C1 ON R.constrid = C1.constrid\n"
    "  JOIN %s\"informix\".SysConstraints C2 ON R.primary = C2.constrid\n"
    "  JOIN %s\"informix\".SysTables T1      ON C1.tabid = T1.tabid\n"
    "  JOIN %s\"informix\".SysTables T2      ON R.ptabid = T2.tabid\n"
    " WHERE C1.constrtype = 'R'\n"
    "   AND %s.tabid = %ld;";

/* Statements for INFO INDEXES -- 5 of them */
/* NB: SysIndexes is a view on SysIndices under IUS or IDS/UDO */
/* One day, it will be OK to use DROP TABLE IF EXISTS _sqlcmd_info_index */
/* But that day is not this day. */
static char info_idx_drop[] = "DROP TABLE _sqlcmd_info_index;";

static char info_idx_mktemp[] =
    "SELECT -- unique explicit indexes\n"
    "       'I' idxtype, I.tabid,"
    " I.owner conowner, I.idxname conname,"
    " I.owner idxowner, I.idxname idxname\n"
    "  FROM %s\"informix\".SysIndexes I\n"
    " WHERE I.tabid = %d AND I.idxname NOT MATCHES ' *' AND i.idxtype = 'U'\n"
    "UNION\n"
    "SELECT -- duplicates explicit indexes\n"
    "       I.idxtype, I.tabid, I.owner conowner, I.idxname conname,"
    " I.owner idxowner, I.idxname idxname\n"
    "  FROM %s\"informix\".SysIndexes I\n"
    " WHERE I.tabid = %d AND I.idxname NOT MATCHES ' *' AND i.idxtype = 'D'\n"
    "UNION\n"
    "SELECT -- primary keys, unique constraints, foreign keys\n"
    "       C.Constrtype idxtype, I.tabid,"
    " C.Owner conowner, C.Constrname conname,"
    " I.Owner idxowner, I.idxname idxname\n"
    "  FROM %s\"informix\".SysIndexes I\n"
    "  JOIN %s\"informix\".SysConstraints C ON C.tabid = I.tabid AND C.Idxname = I.idxname\n"
    " WHERE I.tabid = %d\n"
    " INTO TEMP _sqlcmd_info_index;";

/* Now - about ASC vs DESC on the columns? */
/* And what about the locale for the index? */
static char info_idx_online[] =
    "SELECT t0.idxtype, t0.conowner, t0.conname, t0.idxowner, t0.idxname,\n"
    "       c01.colname col01, CASE WHEN t1.part1 = 0 THEN NULL WHEN t1.part1 < 0 THEN 'D' ELSE 'A' END col01_dir,\n"
    "       c02.colname col02, CASE WHEN t1.part2 = 0 THEN NULL WHEN t1.part2 < 0 THEN 'D' ELSE 'A' END col02_dir,\n"
    "       c03.colname col03, CASE WHEN t1.part3 = 0 THEN NULL WHEN t1.part3 < 0 THEN 'D' ELSE 'A' END col03_dir,\n"
    "       c04.colname col04, CASE WHEN t1.part4 = 0 THEN NULL WHEN t1.part4 < 0 THEN 'D' ELSE 'A' END col04_dir,\n"
    "       c05.colname col05, CASE WHEN t1.part5 = 0 THEN NULL WHEN t1.part5 < 0 THEN 'D' ELSE 'A' END col05_dir,\n"
    "       c06.colname col06, CASE WHEN t1.part6 = 0 THEN NULL WHEN t1.part6 < 0 THEN 'D' ELSE 'A' END col06_dir,\n"
    "       c07.colname col07, CASE WHEN t1.part7 = 0 THEN NULL WHEN t1.part7 < 0 THEN 'D' ELSE 'A' END col07_dir,\n"
    "       c08.colname col08, CASE WHEN t1.part8 = 0 THEN NULL WHEN t1.part8 < 0 THEN 'D' ELSE 'A' END col08_dir,\n"
    "       c09.colname col09, CASE WHEN t1.part9 = 0 THEN NULL WHEN t1.part9 < 0 THEN 'D' ELSE 'A' END col09_dir,\n"
    "       c10.colname col10, CASE WHEN t1.part10 = 0 THEN NULL WHEN t1.part10 < 0 THEN 'D' ELSE 'A' END col10_dir,\n"
    "       c11.colname col11, CASE WHEN t1.part11 = 0 THEN NULL WHEN t1.part11 < 0 THEN 'D' ELSE 'A' END col11_dir,\n"
    "       c12.colname col12, CASE WHEN t1.part12 = 0 THEN NULL WHEN t1.part12 < 0 THEN 'D' ELSE 'A' END col12_dir,\n"
    "       c13.colname col13, CASE WHEN t1.part13 = 0 THEN NULL WHEN t1.part13 < 0 THEN 'D' ELSE 'A' END col13_dir,\n"
    "       c14.colname col14, CASE WHEN t1.part14 = 0 THEN NULL WHEN t1.part14 < 0 THEN 'D' ELSE 'A' END col14_dir,\n"
    "       c15.colname col15, CASE WHEN t1.part15 = 0 THEN NULL WHEN t1.part15 < 0 THEN 'D' ELSE 'A' END col15_dir,\n"
    "       c16.colname col16, CASE WHEN t1.part16 = 0 THEN NULL WHEN t1.part16 < 0 THEN 'D' ELSE 'A' END col16_dir\n"
    "  FROM _sqlcmd_info_index t0\n"
    " INNER JOIN %s\"informix\".SysIndexes t1  ON t0.idxname = t1.idxname AND t0.tabid = t1.tabid\n"
    " INNER JOIN %s\"informix\".SysColumns c01 ON ABS(t1.part1)  = c01.colno AND t1.tabid = c01.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c02 ON ABS(t1.part2)  = c02.colno AND t1.tabid = c02.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c03 ON ABS(t1.part3)  = c03.colno AND t1.tabid = c03.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c04 ON ABS(t1.part4)  = c04.colno AND t1.tabid = c04.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c05 ON ABS(t1.part5)  = c05.colno AND t1.tabid = c05.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c06 ON ABS(t1.part6)  = c06.colno AND t1.tabid = c06.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c07 ON ABS(t1.part7)  = c07.colno AND t1.tabid = c07.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c08 ON ABS(t1.part8)  = c08.colno AND t1.tabid = c08.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c09 ON ABS(t1.part9)  = c09.colno AND t1.tabid = c09.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c10 ON ABS(t1.part10) = c10.colno AND t1.tabid = c10.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c11 ON ABS(t1.part11) = c11.colno AND t1.tabid = c11.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c12 ON ABS(t1.part12) = c12.colno AND t1.tabid = c12.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c13 ON ABS(t1.part13) = c13.colno AND t1.tabid = c13.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c14 ON ABS(t1.part14) = c14.colno AND t1.tabid = c14.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c15 ON ABS(t1.part15) = c15.colno AND t1.tabid = c15.tabid\n"
    "  LEFT JOIN %s\"informix\".SysColumns c16 ON ABS(t1.part16) = c16.colno AND t1.tabid = c16.tabid\n"
    " ORDER BY idxtype, conowner, conname;";

static char info_users[] =
    "SELECT username, usertype FROM \"informix\".SysUsers U\n"
    " WHERE UserType != 'G'\n"
    "   AND (%.*s)\n"
    " ORDER BY username;";

static char info_roles[] =
    "SELECT username, usertype FROM \"informix\".SysUsers U\n"
    " WHERE UserType = 'G'\n"
    "   AND (%.*s)\n"
    " ORDER BY username;";

static char sql_access[] =
    "SELECT ta.grantor, ta.grantee, ta.tabauth, ci.colname, ci.colauth\n"
    "  FROM %s\"informix\".SysTabAuth AS ta\n"
    "  LEFT JOIN (SELECT c.tabid, c.colname, a.colauth, a.grantor, a.grantee\n"
    "               FROM %s\"informix\".SysColauth AS a\n"
    "               JOIN %s\"informix\".SysColumns AS c\n"
    "                 ON a.colno = c.colno AND a.tabid = c.tabid\n"
    "              WHERE c.tabid = %ld\n"
    "            ) AS ci\n"
    "    ON ta.tabid   = ci.tabid\n"
    "   AND ta.grantor = ci.grantor\n"
    "   AND ta.grantee = ci.grantee\n"
    " WHERE ta.tabid   = %ld\n"
    " ORDER BY ta.grantor, ta.grantee, ci.colname;";

/* Statement for INFO STATUS */
static char sql_status[] =
    "SELECT TabID, Tabname, Owner, TabType, RowSize, Ncols, Nrows, Nindexes, Created, Version\n"
    "  FROM %s\"informix\".SysTables T\n"
    " WHERE TabID = %ld;";

/* Statement for INFO FRAGMENTS */
static char sql_fragments[] =
    "SELECT IndexName, Dbspace, FragType, ExprText, Strategy\n"
    "  FROM %s\"informix\".SysFragments F\n"
    " WHERE TabID = %ld\n"
    " ORDER BY IndexName, Dbspace;";

/* Statement for INFO CHECK CONSTRAINTS */
static char sql_checks[] =
    "SELECT co.constrid, constrname, owner, tabid, ck.seqno, ck.checktext\n"
    "  FROM %s\"informix\".SysConstraints co JOIN %s\"informix\".syschecks ck\n"
    "    ON ck.constrid = co.constrid\n"
    " WHERE co.constrtype = 'C'\n"
    "   AND ck.TYPE = 'T'\n"
    "   AND co.TabID = %ld\n"
    " ORDER BY co.Constrname, co.Owner, ck.seqno;";

/* Deal with simple INFO queries with single replacement for WHERE clause */
static void info_where(const char *sql_fmt, const char *wh_start, const char *wh_end)
{
    int wh_len = (wh_start) ? (wh_end - wh_start) : sizeof("1=1");
    int sql_len = strlen(sql_fmt) + wh_len;
    char *sql = (char *)MALLOC(sql_len);    /*=C++=*/
    const char *wh = wh_start ? wh_start : "1=1";
    DB_TRACKING();
    esnprintf(sql, sql_len, sql_fmt, wh_len, wh);
    (*sql_exec)(sql);
    FREE(sql);
}

/*
** Convert connection information into full table name.
** NB: quotes or absence of quotes around owner name are preserved by
** grammar so that MODE ANSI databases work correctly.  Hence, do not
** enclose owner in quotes here; it would defeat the good work done
** by the grammar.
*/

static void info_viewbody(InfoInfo *info)
{
    char buffer[1024];
    long viewid;
    char dbase[300];

    viewid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (viewid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_viewbody, dbase, viewid);
        DB_TRACE(3, "info_viewbody: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

static void info_trigbody(InfoInfo *info)
{
    char buffer[1024];
    long trigid;
    char dbase[300];

    trigid = sql_trigid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (trigid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_trigbody, dbase, trigid);
        DB_TRACE(3, "info_trigbody: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

static void info_procbody(InfoInfo *info)
{
    char buffer[1024];
    long procid;
    char dbase[300];

    procid = sql_procid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (procid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_procbody, dbase, procid);
        DB_TRACE(3, "info_procbody: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

/*
** Convert connection information into full table name.
** NB: quotes or absence of quotes around owner name are preserved by
** grammar so that MODE ANSI databases work correctly.  Hence, do not
** enclose owner in quotes here; it would defeat the good work done
** by the grammar.
*/

static void info_columns(InfoInfo *info)
{
    char buffer[1024];
    long tabid;
    char dbase[300];

    DB_TRACE(3, "info_columns: owner = <<%s>>\n", info->owner);
    DB_TRACE(3, "info_columns: table = <<%s>>\n", info->table);
    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_columns, dbase, tabid);
        DB_TRACE(3, "info_columns: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

/*
** Convert connection information into full table name.
** NB: quotes or absence of quotes around owner name are preserved by
** grammar so that MODE ANSI databases work correctly.  Hence, do not
** enclose owner in quotes here; it would defeat the good work done
** by the grammar.
*/

static void info_refs(InfoInfo *info, const char *alias, const char *function)
{
    char buffer[1024];
    long tabid;
    char dbase[300];

    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_references, dbase, dbase, dbase, dbase, dbase, alias, tabid);
        DB_TRACE(3, "%s: %s\n", function, buffer);
        (*sql_exec)(buffer);
    }
}

static void info_refs_by(InfoInfo *info)
{
    info_refs(info, "T1", __func__);
}

static void info_refs_to(InfoInfo *info)
{
    info_refs(info, "T2", __func__);
}

static void info_access(InfoInfo *info)
{
    char buffer[1024];
    long tabid;
    char dbase[300];

    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_access, dbase, dbase, dbase, tabid, tabid);
        DB_TRACE(3, "info_access: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

static void info_status(InfoInfo *info)
{
    char buffer[1024];
    long tabid;
    char dbase[300];

    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_status, dbase, tabid);
        DB_TRACE(3, "info_status: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

static void info_fragments(InfoInfo *info)
{
    char buffer[1024];
    long tabid;
    char dbase[300];

    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_fragments, dbase, tabid);
        DB_TRACE(3, "info_fragments: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

static void info_checks(InfoInfo *info)
{
    char buffer[1024];
    long tabid;
    char dbase[300];

    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_checks, dbase, dbase, tabid);
        DB_TRACE(3, "info_checks: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

static void info_triggers(InfoInfo *info)
{
    char buffer[1024];
    long tabid;
    char dbase[300];

    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), sql_triggers, dbase, dbase, tabid, tabid);
        DB_TRACE(3, "info_triggers: %s\n", buffer);
        (*sql_exec)(buffer);
    }
}

static void drop_info_idx_temp(void)
{
    ctxt_setcontinue(op_push);
    ctxt_setcontinue(op_on);
    ctxt_setsilence(op_push);
    ctxt_setsilence(op_on);
    (*sql_exec)(info_idx_drop);
    ctxt_setsilence(op_pop);
    ctxt_setcontinue(op_pop);
}

static void info_indexes(InfoInfo * info)
{
    EXEC SQL BEGIN DECLARE SECTION;
    char buffer[8196];
    EXEC SQL END DECLARE SECTION;
    long tabid;
    char dbase[300];

    tabid = sql_tabid(info->table, info->owner, info->dbase, info->server,
                        conninfo_current_modeansi());

    if (tabid < 0)
        sql_error();
    else
    {
        ctxt_newcontext();
        drop_info_idx_temp();

        sql_mkdbasename(info->dbase, info->server, dbase, sizeof(dbase));
        esnprintf(buffer, sizeof(buffer), info_idx_mktemp, dbase, tabid, dbase, tabid, dbase,
                dbase, tabid);
        DB_TRACE(3, "info_indexes: %s\n", buffer);
        (*sql_exec)(buffer);
        esnprintf(buffer, sizeof(buffer), info_idx_online, dbase, dbase, dbase, dbase, dbase,
                dbase, dbase, dbase, dbase, dbase, dbase, dbase, dbase,
                dbase, dbase, dbase, dbase);
        DB_TRACE(3, "info_indexes: %s\n", buffer);
        (*sql_exec)(buffer);
        drop_info_idx_temp();
        ctxt_endcontext();
    }
}

void     sql_info(char *str)
{
    InfoInfo stmt;

    switch (parse_infostmt(str, &stmt))
    {
    case INFO_TABLES:
        info_where(info_tables, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_SYSTABLES:
        info_where(info_systables, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_COLUMNS:
        info_columns(&stmt);
        break;

    case INFO_PROCBODY:
        info_procbody(&stmt);
        break;

    case INFO_PROCEDURES:
        info_where(info_procedures, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_VIEWS:
        info_where(info_views, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_VIEWBODY:
        info_viewbody(&stmt);
        break;

    case INFO_BASETABLES:
        info_where(info_basetables, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_SYNONYMS:
        info_where(info_synonyms, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_INDEXES:
        info_indexes(&stmt);
        break;

    case INFO_ACCESS: /* aka PRIVILEGES */
        info_access(&stmt);
        break;

    case INFO_DATABASES:
        sql_dbnames();
        break;

    case INFO_CONNECTIONS:
        conninfo_print();
        break;

    case INFO_REFS_BY:
        info_refs_by(&stmt);
        break;

    case INFO_REFS_TO:
        info_refs_to(&stmt);
        break;

    case INFO_STATUS:
        info_status(&stmt);
        break;

    case INFO_FRAGMENTS:
        info_fragments(&stmt);
        break;

    case INFO_CONSTR_CHECK:
        info_checks(&stmt);
        break;

    case INFO_TRIGGERS:
        info_triggers(&stmt);
        break;

    case INFO_TRIGBODY:
        info_trigbody(&stmt);
        break;

    case INFO_USERS:
        info_where(info_users, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_ROLES:
        info_where(info_roles, stmt.wh_start, stmt.wh_end);
        break;

    case INFO_HELP:
        fprintf(ctxt_output(),
                "-- Help for the INFO command:\n"
                "-- INFO ACCESS FOR table (aka INFO PRIVILEGES)\n"
                "-- INFO BASETABLES [WHERE ...]\n"
                "-- INFO CHECK CONSTRAINTS FOR table\n"
                "-- INFO COLUMNS FOR table\n"
                "-- INFO CONNECTIONS\n"
                "-- INFO DATABASES\n"
                "-- INFO FRAGMENTS FOR table\n"
                "-- INFO HELP\n"
                "-- INFO INDEXES FOR table\n"
                "-- INFO PROCBODY FOR procname\n"
                "-- INFO PROCEDURES [WHERE ...]\n"
                "-- INFO REFERENCES BY table\n"
                "-- INFO REFERENCES TO table\n"
                "-- INFO ROLES [WHERE ...]\n"
                "-- INFO STATUS FOR table\n"
                "-- INFO SYNONYMS [WHERE ...]\n"
                "-- INFO SYSTABLES [WHERE ...]\n"
                "-- INFO TABLES [WHERE ...]\n"
                "-- INFO TRIGBODY FOR trigger\n"
                "-- INFO TRIGGERS FOR table\n"
                "-- INFO USERS [WHERE ...]\n"
                "-- INFO VIEWBODY for view\n"
                "-- INFO VIEWS [WHERE ...]\n"
                );
        break;

    case INFO_ERROR:
        error_746("Syntax error in INFO statement");
        break;

    default:
        error_746("Internal error handling INFO statement");
        break;
    }
}
